/* cue_info.c - Cue file support for XMMS
 * 
 * Copyright (C) 2003  Oskar Liljeblad
 *
 * This file is part of xmms-cueinfo.
 *
 * This software is copyrighted work licensed under the terms of the
 * GNU General Public License. Please consult the file "COPYING" for
 * details.
 */

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>

#define MAX_CUE_LINE_LENGTH 1000
#define MAX_CUE_TRACKS 1000

#define EMPTIZE(x) ((x)==NULL ? "":(x))

static void init(void);
static void cleanup(void);
static gint timeout_func(gpointer);
static void free_cue_data(void);
static void cache_cue_file(FILE *file);
static void fix_cue_argument(char *line);
static void entry_changed(GtkWidget *entry);

static GtkWidget *info_win;
static GtkWidget *track_spin;
static GtkWidget *performer_entry;
static GtkWidget *title_entry;
static gchar *cue_performer = NULL;
static gchar *cue_title = NULL;
static gint last_cue_track = 0;
static gint cur_cue_track = 0;
static gint entry_lock = 0;
static struct {
	gchar *performer;
	gchar *title;
	gint index;
} cue_tracks[MAX_CUE_TRACKS];
static gint previous_song = -1;
static gint previous_length = -2;
static gint timeout_tag = 0;

static GeneralPlugin cue_gp =
{
	NULL,			/* handle */
	NULL,			/* filename */
	-1,				/* xmms_session */
	NULL,			/* description */
	init,	       	/* init */
	NULL,	       	/* about */
	NULL,	  	   	/* configure */
	cleanup,       	/* cleanup */
};

GeneralPlugin *get_gplugin_info(void)
{
	cue_gp.description = g_strdup_printf("Cue Info %s", VERSION);
	return &cue_gp;
}

static gboolean window_delete(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gtk_widget_hide(widget);
	return TRUE;
}

static void init(void)
{
	GtkWidget *table;
	GtkWidget *label;

	info_win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(info_win), "XMMS Cue Info");
	table = gtk_table_new(2, 3, FALSE);

	label = gtk_label_new("Track: ");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

	label = gtk_label_new("Performer: ");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

	label = gtk_label_new("Title: ");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 0, 0);

	track_spin = gtk_spin_button_new(NULL, 1, 0);
	gtk_signal_connect(GTK_OBJECT(track_spin), "changed", entry_changed, NULL);
	gtk_table_attach(GTK_TABLE(table), track_spin, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

	performer_entry = gtk_entry_new();
	gtk_entry_set_editable(GTK_ENTRY(performer_entry), FALSE);
	gtk_table_attach(GTK_TABLE(table), performer_entry, 1, 2, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL, 0, 0);

	title_entry = gtk_entry_new();
	gtk_entry_set_editable(GTK_ENTRY(title_entry), FALSE);
	gtk_table_attach(GTK_TABLE(table), title_entry, 1, 2, 2, 3, GTK_FILL|GTK_EXPAND, GTK_FILL, 0, 0);

	gtk_container_add(GTK_CONTAINER(info_win), table);
	gtk_container_set_border_width(GTK_CONTAINER(info_win), 6);

	gtk_signal_connect(GTK_OBJECT(info_win), "delete_event", GTK_SIGNAL_FUNC(window_delete), NULL);

	previous_song = -1;
	previous_length = -2;
	timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
}

static void cleanup(void)
{
	if (timeout_tag)
		gtk_timeout_remove(timeout_tag);
	timeout_tag = 0;
	free_cue_data();
}

static void entry_changed(GtkWidget *entry)
{
	gint new_cue_track;

	if (entry_lock)
		return;

	new_cue_track = atoi(gtk_entry_get_text(GTK_ENTRY(entry)));
	if (new_cue_track >= 1 && new_cue_track <= last_cue_track)
	{
		gint offset;
	
		/* For some reason, when you use jump_to_time, xmms is sometimes
		 * off (early) by up to 1 second. That's why we compensate by adding
		 * 1000.
		 */
		offset = (new_cue_track == 1 ? 0 : 1000);
		xmms_remote_jump_to_time(cue_gp.xmms_session, cue_tracks[new_cue_track-1].index + offset);
	}
}

static gint timeout_func(gpointer data)
{
	gboolean running;
	gboolean show_hide_win = FALSE;

	GDK_THREADS_ENTER();

	running = xmms_remote_is_running(cue_gp.xmms_session);

	if (running)
	{
		gint pos;
		gint length;
		gint time;
		gint prev_cue_track;

		pos = xmms_remote_get_playlist_pos(cue_gp.xmms_session);
		length = xmms_remote_get_playlist_time(cue_gp.xmms_session, pos);

		if (pos != previous_song || length != previous_length)
		{
			gchar *filename;
			gint filelength;

			previous_song = pos;
			previous_length = length;

			free_cue_data();

			filename = xmms_remote_get_playlist_file(cue_gp.xmms_session, pos);
			show_hide_win = TRUE;
			if (filename != NULL) {
				filelength = strlen(filename);
				if (filelength >= 4)
				{
					FILE *file;

					strcpy(&filename[filelength-3], "cue");
					file = fopen(filename, "r");
					if (file != NULL) {
						cache_cue_file(file);
						fclose(file);
					}
				}
				cur_cue_track = 0;

				if (last_cue_track > 0) {
					GtkObject *adjustment;

					adjustment = gtk_adjustment_new(1, 1, last_cue_track, 1, 2, 2);
					gtk_spin_button_set_adjustment(GTK_SPIN_BUTTON(track_spin), GTK_ADJUSTMENT(adjustment));
				}

				g_free(filename);
			}
		}

		time = xmms_remote_get_output_time(cue_gp.xmms_session);
		prev_cue_track = cur_cue_track;

		if (cur_cue_track == 0)
			cur_cue_track = 1;

		for (; cur_cue_track < last_cue_track; cur_cue_track++) {
			if (cue_tracks[cur_cue_track-1].index <= time
					&& (cur_cue_track == last_cue_track
					|| time < cue_tracks[cur_cue_track].index))
				break;
		}
		for (; cur_cue_track > 1; cur_cue_track--) {
			if (cue_tracks[cur_cue_track-1].index <= time
					&& (cur_cue_track == last_cue_track
					|| time < cue_tracks[cur_cue_track].index))
				break;
		}

		if (prev_cue_track != cur_cue_track)
		{
			entry_lock++;
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(track_spin), cur_cue_track);
			gtk_entry_set_text(GTK_ENTRY(performer_entry), EMPTIZE(cue_tracks[cur_cue_track-1].performer));
			gtk_entry_set_text(GTK_ENTRY(title_entry), EMPTIZE(cue_tracks[cur_cue_track-1].title));
			entry_lock--;
		}
	}

	GDK_THREADS_LEAVE();

	if (show_hide_win)
	{
		if (last_cue_track > 0)
			gtk_widget_show_all(info_win);
		else
			gtk_widget_hide(info_win);
	}

	return TRUE;
}

static void free_cue_data(void)
{
	g_free(cue_performer);
	cue_performer = NULL;
	g_free(cue_title);
	cue_title = NULL;
	for (; last_cue_track > 0; last_cue_track--) {
		g_free(cue_tracks[last_cue_track-1].performer);
		g_free(cue_tracks[last_cue_track-1].title);
	}
}

static void cache_cue_file(FILE *file)
{
	gchar line[MAX_CUE_LINE_LENGTH+1];

	while (TRUE) {
		gint p;
		gint q;

		if (fgets(line, MAX_CUE_LINE_LENGTH+1, file) == NULL)
			return;

		for (p = 0; line[p] && isspace(line[p]); p++);
		if (!line[p])
			continue;
		for (q = p; line[q] && !isspace(line[q]); q++);
		if (!line[q])
			continue;
		line[q] = '\0';
		for (q++; line[q] && isspace(line[q]); q++);

		if (strcasecmp(line+p, "PERFORMER") == 0) {
			fix_cue_argument(line+q);
			if (last_cue_track == 0) {
				cue_performer = g_strdup(line+q);
			} else {
				cue_tracks[last_cue_track-1].performer = g_strdup(line+q);
			}
		}
		else if (strcasecmp(line+p, "TITLE") == 0) {
			fix_cue_argument(line+q);
			if (last_cue_track == 0) {
				cue_title = g_strdup(line+q);
			} else {
				cue_tracks[last_cue_track-1].title = g_strdup(line+q);
			}
		}
		else if (strcasecmp(line+p, "TRACK") == 0) {
			gint track;

			fix_cue_argument(line+q);
			for (p = q; line[p] && isdigit(line[p]); p++);
			line[p] = '\0';
			for (; line[q] && line[q] == '0'; q++);
			if (!line[q])
				continue;
			track = atoi(line+q);
			if (track >= MAX_CUE_TRACKS)
				continue;
			last_cue_track = track;
			cue_tracks[last_cue_track-1].index = 0;
			cue_tracks[last_cue_track-1].performer = NULL;
			cue_tracks[last_cue_track-1].title = NULL;
		}
		else if (strcasecmp(line+p, "INDEX") == 0) {
			for (p = q; line[p] && !isspace(line[p]); p++);
			if (!line[p])
				continue;
			for (p++; line[p] && isspace(line[p]); p++);
			for (q = p; line[q] && !isspace(line[q]); q++);
			if (q-p >= 8 && line[p+2] == ':' && line[p+5] == ':') {
				cue_tracks[last_cue_track-1].index =
						((line[p+0]-'0')*10 + (line[p+1]-'0')) * 60000 +
						((line[p+3]-'0')*10 + (line[p+4]-'0')) * 1000 +
						((line[p+6]-'0')*10 + (line[p+7]-'0')) * 10;
			}
		}
	}
}

static void fix_cue_argument(char *line)
{
	if (line[0] == '"') {
		gchar *l2;
		for (l2 = line+1; *l2 && *l2 != '"'; l2++)
				*(l2-1) = *l2;
			*(l2-1) = *l2;
		for (; *line && *line != '"'; line++) {
			if (*line == '\\' && *(line+1)) {
				for (l2 = line+1; *l2 && *l2 != '"'; l2++)
					*(l2-1) = *l2;
				*(l2-1) = *l2;
			}
		}
		*line = '\0';
	}
	else {
		for (; *line && *line != '\r' && *line != '\n'; line++);
		*line = '\0';
	}
}
